
/*
*  jobctl.c
*
*  run a program and wait for it to finish, while taking job control
*  commands over a FIFO.
*
*  2003-10-3 manchek
*/

#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>

#define FIFOPATH "/tmp/jobctl"

int pfd[2];     /* pipe to collect exited child status in main loop */

/* reap()
*
*  reap any exited child status (without blocking) and stuff the
*  results of wait into the collection pipe.
*/
void
reap(
    int sig )
{
    int x[2];

    x[1] = -1;
    x[0] = waitpid(-1, &x[1], WNOHANG);
    write(pfd[1], (char *)&x, sizeof(x));
}

#define ACTION_NONE 0
#define ACTION_RESTART 1

int
main(
    int argc,
    char **argv )
{
    int pid;                    /* child pid */
    int action = ACTION_NONE;   /* restart action */
    int xs = 0;                 /* exit status */
    int a;
    char *fifopath = FIFOPATH;
    int errs = 0;

    a = 1;
    while (a < argc && argv[a][0] == '-') {
	if (!strcmp("-?", argv[a])
	|| !strcmp("-h", argv[a])
	|| !strcmp("-help", argv[a])) {
	    fprintf(stderr, "usage: jobctl [ options ] program args ...\n");
	    fprintf(stderr, "   -fifo path  FIFO filename (default %s)\n",
		    FIFOPATH);
/*
	    fprintf(stderr, "   -force      recreate FIFO if necessary\n");
	    fprintf(stderr, "   -remove     remove FIFO on exit\n");
	    fprintf(stderr, "   -private    set permissions for owner only\n");
	    fprintf(stderr, "   -public     set permissions for all users\n");
*/
/*
	    fprintf(stderr, "options to send command to running jobctl:\n");
	    fprintf(stderr, "   -send cmd   command to send\n");
	    fprintf(stderr, "   -tout src   timeout for sending (default forever)\n");
*/
	    exit(0);

	} else if (!strcmp("-fifo", argv[a])) {
	    if (a == argc - 1) {
		fprintf(stderr, "missing parameter for -fifo\n");
		errs++;

	    } else {
		a++;
		fifopath = argv[a];
	    }

	} else {
	    fprintf(stderr, "unknown option: %s\n", argv[a]);
	    errs++;
	}
	a++;
    }

    if (errs || argc - a < 1) {
	fprintf(stderr, "usage: jobctl [ options ] program args ...\n");
	exit(2);
    }

    if (mkfifo(fifopath, 0666)) {
	struct stat sb;

	if (errno != EEXIST) {
	    perror("mkfifo");
	    exit(1);
	}
	if (stat(fifopath, &sb)) {
	    perror("stat");
	    exit(1);
	}
	if (!S_ISFIFO(sb.st_mode)) {
	    if (unlink(fifopath)) {
		perror("unlink");
		exit(1);
	    }
	    if (mkfifo(fifopath, 0666)) {
		perror("unlink");
		exit(1);
	    }
	}
    }
    if (chmod(fifopath, 0666))
	perror("chmod");

    if (pipe(pfd)) {
	perror("pipe");
	exit(1);
    }
    signal(SIGCHLD, reap);

restart:

    switch ((pid = fork())) {
    case -1:
	perror("fork");
	exit(1);
	break;

    case 0:
	close(pfd[0]);
	close(pfd[1]);
	execv(argv[a], argv + a);
	perror(argv[a]);
	exit(3);
	break;

    default:
	{
	    int ffd;
	    int nfds;
	    int s;
	    fd_set fds;

	    ffd = open(fifopath, O_NONBLOCK|O_RDONLY, 0);
#ifdef DEBUG
	    fprintf(stderr, "got fifo %d\n", ffd);
#endif
	    nfds = (ffd > pfd[0] ? ffd : pfd[0]) + 1;
	    while (1) {
		FD_ZERO(&fds);
		FD_SET(ffd, &fds);
		FD_SET(pfd[0], &fds);
		s = select(nfds, &fds, (fd_set *)0, (fd_set *)0,
			(struct timeval *)0);
#ifdef DEBUG
		fprintf(stderr, "select ret %d\n", s);
#endif
		if (s > 0) {
		    if (FD_ISSET(pfd[0], &fds)) {
			int x[2];
			read(pfd[0], (char *)&x, sizeof(x));
			if (x[0] == pid) {
#ifdef DEBUG
			    fprintf(stderr, "child exited with %d\n", x[1]);
#endif
			    xs = WEXITSTATUS(x[1]);
			    break;
			}
		    }
		    if (FD_ISSET(ffd, &fds)) {
			int n;
			char buf[40];

			if ((n = read(ffd, buf, sizeof(buf))) == -1)
			    perror("read");
			else {
			    buf[n] = 0;
#ifdef DEBUG
			    fprintf(stderr, "CMD %d <%s>\n", n, buf);
#endif
			}
			close(ffd);
			ffd = open(fifopath, O_NONBLOCK|O_RDONLY, 0);
#ifdef DEBUG
			fprintf(stderr, "got fifo %d\n", ffd);
#endif
			nfds = (ffd > pfd[0] ? ffd : pfd[0]) + 1;

			if (n > 0) {
			    if (!strcmp("kill\n", buf)) {
				kill(pid, SIGTERM);

			    } else if (!strcmp("restart\n", buf)) {
				action = ACTION_RESTART;
				kill(pid, SIGTERM);

			    } else if (!strcmp("pause\n", buf)) {
				kill(pid, SIGSTOP);

			    } else if (!strcmp("resume\n", buf)) {
				kill(pid, SIGCONT);
			    }
			}
		    }
		}
	    }
	    close(ffd);
	}
	break;
    }

    if (action == ACTION_RESTART) {
	action = ACTION_NONE;
	goto restart;
    }

    exit(xs);
}

